@@ -1,12 +1,12 @@
 """ Schema Tester """
+from __future__ import annotations
+
 from itertools import chain
-from typing import Any, Callable, Dict, List, Optional, Union, cast
+from typing import TYPE_CHECKING, Any, Callable, List, Optional, cast
 
 from django.conf import settings
 from django.core.exceptions import ImproperlyConfigured
-from rest_framework.response import Response
 
-from openapi_tester import type_declarations as td
 from openapi_tester.constants import (
     INIT_ERROR,
     UNDOCUMENTED_SCHEMA_SECTION_ERROR,
@@ -37,20 +37,24 @@ from openapi_tester.validators import (
     validate_unique_items,
 )
 
+if TYPE_CHECKING:
+
+    from rest_framework.response import Response
+
 
 class SchemaTester:
     """Schema Tester: this is the base class of the library."""
 
-    loader: Union[StaticSchemaLoader, DrfSpectacularSchemaLoader, DrfYasgSchemaLoader]
-    validators: List[Callable[[dict, Any], Optional[str]]]
+    loader: StaticSchemaLoader | DrfSpectacularSchemaLoader | DrfYasgSchemaLoader
+    validators: list[Callable[[dict, Any], str | None]]
 
     def __init__(
         self,
-        case_tester: Optional[Callable[[str], None]] = None,
-        ignore_case: Optional[List[str]] = None,
-        schema_file_path: Optional[str] = None,
-        validators: Optional[List[Callable[[dict, Any], Optional[str]]]] = None,
-        field_key_map: Optional[Dict[str, str]] = None,
+        case_tester: Callable[[str], None] | None = None,
+        ignore_case: list[str] | None = None,
+        schema_file_path: str | None = None,
+        validators: list[Callable[[dict, Any], str | None]] | None = None,
+        field_key_map: dict[str, str] | None = None,
     ) -> None:
         """
         Iterates through an OpenAPI schema object and API response to check that they match at every level.
@@ -74,7 +78,7 @@ class SchemaTester:
             raise ImproperlyConfigured(INIT_ERROR)
 
     @staticmethod
-    def get_key_value(schema: dict, key: str, error_addon: str = "") -> dict:
+    def get_key_value(schema: dict[str, dict], key: str, error_addon: str = "") -> dict:
         """
         Returns the value of a given key
         """
@@ -86,7 +90,7 @@ class SchemaTester:
             ) from e
 
     @staticmethod
-    def get_status_code(schema: dict, status_code: Union[str, int], error_addon: str = "") -> dict:
+    def get_status_code(schema: dict[str | int, dict], status_code: str | int, error_addon: str = "") -> dict:
         """
         Returns the status code section of a schema, handles both str and int status codes
         """
@@ -99,14 +103,14 @@ class SchemaTester:
         )
 
     @staticmethod
-    def get_schema_type(schema: dict) -> Optional[str]:
+    def get_schema_type(schema: dict[str, str]) -> str | None:
         if "type" in schema:
             return schema["type"]
         if "properties" in schema or "additionalProperties" in schema:
             return "object"
         return None
 
-    def get_response_schema_section(self, response: td.Response) -> Dict[str, Any]:
+    def get_response_schema_section(self, response: Response) -> dict[str, Any]:
         """
         Fetches the response section of a schema, wrt. the route, method, status code, and schema version.
 
@@ -127,14 +131,16 @@ class SchemaTester:
         method_object = self.get_key_value(
             route_object,
             response_method,
-            f"\n\nUndocumented method: {response_method}.\n\nDocumented methods: {[method.lower() for method in route_object.keys() if method.lower() != 'parameters']}.",
+            f"\n\nUndocumented method: {response_method}.\n\nDocumented methods: "
+            f"{[method.lower() for method in route_object.keys() if method.lower() != 'parameters']}.",
         )
 
         responses_object = self.get_key_value(method_object, "responses")
         status_code_object = self.get_status_code(
             responses_object,
             response.status_code,
-            f"\n\nUndocumented status code: {response.status_code}.\n\nDocumented status codes: {list(responses_object.keys())}. ",
+            f"\n\nUndocumented status code: {response.status_code}.\n\n"
+            f"Documented status codes: {list(responses_object.keys())}. ",
         )
 
         if "openapi" not in schema:  # pylint: disable=E1135
@@ -150,32 +156,42 @@ class SchemaTester:
             json_object = self.get_key_value(
                 content_object,
                 "application/json",
-                f"\n\nNo `application/json` responses documented for method: {response_method}, path: {parameterized_path}",
+                "\n\nNo `application/json` responses documented for method: "
+                f"{response_method}, path: {parameterized_path}",
             )
             return self.get_key_value(json_object, "schema")
 
-        if response.json():
+        if response.data and response.json():
             raise UndocumentedSchemaSectionError(
                 UNDOCUMENTED_SCHEMA_SECTION_ERROR.format(
                     key="content",
-                    error_addon=f"\n\nNo `content` defined for this response: {response_method}, path: {parameterized_path}",
+                    error_addon=(
+                        f"\n\nNo `content` defined for this response: {response_method}, path: {parameterized_path}"
+                    ),
                 )
             )
         return {}
 
-    def handle_one_of(self, schema_section: dict, data: Any, reference: str, **kwargs: Any):
+    def handle_one_of(self, schema_section: dict, data: Any, reference: str, **kwargs: Any) -> None:
         matches = 0
+        passed_schema_section_formats = set()
         for option in schema_section["oneOf"]:
             try:
                 self.test_schema_section(schema_section=option, data=data, reference=f"{reference}.oneOf", **kwargs)
                 matches += 1
+                passed_schema_section_formats.add(option.get("format"))
             except DocumentationError:
                 continue
+        if matches == 2 and passed_schema_section_formats == {"date", "date-time"}:
+            # With Django v4, the datetime validator now parses normal
+            # date formats successfully, so a oneOf: date // datetime section
+            # will succeed twice where it used to succeed once.
+            return
         if matches != 1:
             raise DocumentationError(f"{VALIDATE_ONE_OF_ERROR.format(matches=matches)}\n\nReference: {reference}.oneOf")
 
-    def handle_any_of(self, schema_section: dict, data: Any, reference: str, **kwargs: Any):
-        any_of: List[Dict[str, Any]] = schema_section.get("anyOf", [])
+    def handle_any_of(self, schema_section: dict, data: Any, reference: str, **kwargs: Any) -> None:
+        any_of: list[dict[str, Any]] = schema_section.get("anyOf", [])
         for schema in chain(any_of, lazy_combinations(any_of)):
             try:
                 self.test_schema_section(schema_section=schema, data=data, reference=f"{reference}.anyOf", **kwargs)
@@ -205,8 +221,8 @@ class SchemaTester:
     def test_key_casing(
         self,
         key: str,
-        case_tester: Optional[Callable[[str], None]] = None,
-        ignore_case: Optional[List[str]] = None,
+        case_tester: Callable[[str], None] | None = None,
+        ignore_case: list[str] | None = None,
     ) -> None:
         tester = case_tester or getattr(self, "case_tester", None)
         ignore_case = [*self.ignore_case, *(ignore_case or [])]
@@ -218,7 +234,7 @@ class SchemaTester:
         schema_section: dict,
         data: Any,
         reference: str = "init",
-        validators: Optional[List[Callable[[dict, dict], Optional[str]]]] = None,
+        validators: list[Callable[[dict, dict], str | None]] | None = None,
         **kwargs: Any,
     ) -> None:
         """
@@ -231,7 +247,7 @@ class SchemaTester:
             raise DocumentationError(
                 f"{VALIDATE_NONE_ERROR}\n\n"
                 f"Reference: {reference}\n\n"
-                f"Hint: Return a valid type, or document the value as nullable"
+                "Hint: Return a valid type, or document the value as nullable"
             )
         schema_section = normalize_schema_section(schema_section)
         if "oneOf" in schema_section:
@@ -245,7 +261,7 @@ class SchemaTester:
         if not schema_section_type:
             return
         combined_validators = cast(
-            List[Callable],
+            List[Callable[[dict, Any], Optional[str]]],
             [
                 validate_type,
                 validate_format,
@@ -280,8 +296,8 @@ class SchemaTester:
         schema_section: dict,
         data: dict,
         reference: str,
-        case_tester: Optional[Callable[[str], None]] = None,
-        ignore_case: Optional[List[str]] = None,
+        case_tester: Callable[[str], None] | None = None,
+        ignore_case: list[str] | None = None,
     ) -> None:
         """
         1. Validate that casing is correct for both response and schema
@@ -294,7 +310,7 @@ class SchemaTester:
         write_only_properties = [key for key in properties.keys() if properties[key].get("writeOnly")]
         required_keys = [key for key in schema_section.get("required", []) if key not in write_only_properties]
         response_keys = data.keys()
-        additional_properties: Optional[Union[bool, dict]] = schema_section.get("additionalProperties")
+        additional_properties: bool | dict | None = schema_section.get("additionalProperties")
         additional_properties_allowed = additional_properties is not None
         if additional_properties_allowed and not isinstance(additional_properties, (bool, dict)):
             raise OpenAPISchemaError("Invalid additionalProperties type")
@@ -304,7 +320,7 @@ class SchemaTester:
                 raise DocumentationError(
                     f"{VALIDATE_MISSING_RESPONSE_KEY_ERROR.format(missing_key=key)}\n\nReference: {reference}."
                     f"object:key:{key}\n\nHint: Remove the key from your"
-                    f" OpenAPI docs, or include it in your API response"
+                    " OpenAPI docs, or include it in your API response"
                 )
         for key in response_keys:
             self.test_key_casing(key, case_tester, ignore_case)
@@ -317,7 +333,7 @@ class SchemaTester:
                 raise DocumentationError(
                     f"{VALIDATE_WRITE_ONLY_RESPONSE_KEY_ERROR.format(write_only_key=key)}\n\nReference: {reference}"
                     f".object:key:{key}\n\nHint: Remove the key from your API response, or remove the "
-                    f'"WriteOnly" restriction'
+                    '"WriteOnly" restriction'
                 )
         for key, value in data.items():
             if key in properties:
@@ -337,7 +353,7 @@ class SchemaTester:
                     ignore_case=ignore_case,
                 )
 
-    def test_openapi_array(self, schema_section: dict, data: dict, reference: str, **kwargs: Any) -> None:
+    def test_openapi_array(self, schema_section: dict[str, Any], data: dict, reference: str, **kwargs: Any) -> None:
         for datum in data:
             self.test_schema_section(
                 # the items keyword is required in arrays
@@ -350,10 +366,10 @@ class SchemaTester:
     def validate_response(
         self,
         response: Response,
-        case_tester: Optional[Callable[[str], None]] = None,
-        ignore_case: Optional[List[str]] = None,
-        validators: Optional[List[Callable[[dict, Any], Optional[str]]]] = None,
-    ):
+        case_tester: Callable[[str], None] | None = None,
+        ignore_case: list[str] | None = None,
+        validators: list[Callable[[dict[str, Any], Any], str | None]] | None = None,
+    ) -> None:
         """
         Verifies that an OpenAPI schema definition matches an API response.
 
@@ -367,7 +383,7 @@ class SchemaTester:
         response_schema = self.get_response_schema_section(response)
         self.test_schema_section(
             schema_section=response_schema,
-            data=response.json(),
+            data=response.json() if response.data else {},
             case_tester=case_tester or self.case_tester,
             ignore_case=ignore_case,
             validators=validators,
